"""Huggingface Dataset Handler."""

from typing import TYPE_CHECKING, Any

import pandas as pd

if TYPE_CHECKING:
    from datasets import Dataset


class HuggingfaceDatasetHandler:
    """Handler for Huggingface Datasets."""

    def __init__(
        self, dataset_name: str, dataset: dict | list | pd.DataFrame, dataset_split: str = "train"
    ) -> None:
        self.dataset_name = dataset_name
        self.dataset = self.transform_to_dataset(dataset)
        self.dataset_split = dataset_split

    def transform_to_dataset(self, dataset: dict | list | pd.DataFrame | Any) -> "Dataset":
        """Transform the data to a Huggingface Dataset."""
        from datasets import Dataset

        if isinstance(dataset, list):
            return Dataset.from_list(dataset)
        elif isinstance(dataset, dict):
            return Dataset.from_dict(dataset)
        elif isinstance(dataset, pd.DataFrame):
            return Dataset.from_pandas(dataset)
        else:
            raise ValueError("Dataset must be a list, dictionary, or pandas DataFrame.")

    def push_to_hf(self, private: bool = True, **kwargs: Any) -> None:
        """Push the dataset to the Huggingface Hub."""
        self.dataset.push_to_hub(
            repo_id=self.dataset_name, split=self.dataset_split, private=private, **kwargs
        )



"""Factory for creating RAG pipeline instances."""

from typing import TYPE_CHECKING, Type

from encourage.rag.base.config import BaseRAGConfig
from encourage.rag.base.enum import RAGMethod

if TYPE_CHECKING:
    from encourage.rag.base_impl import BaseRAG


class RAGFactory:
    """Factory to create RAG pipeline instances from registered RAG classes and configs.

    It uses the Strategy and Factory patterns to dynamically register and instantiate
    various RAG implementations using a decorator-based registry.
    """

    # Registry mapping method name → (config class, RAG class)
    registry: dict[RAGMethod, tuple[Type[BaseRAGConfig], Type["BaseRAG"]]] = {}

    @classmethod
    def register(cls, name: RAGMethod, config_cls: Type[BaseRAGConfig]):
        """Register a new RAG method to the factory.

        Args:
            name: Unique string identifier for the RAG method.
            config_cls: Pydantic config class for validating initialization args.

        Returns:
            A decorator that registers the corresponding RAG class.

        """

        def decorator(rag_cls: Type["BaseRAG"]):
            cls.registry[name] = (config_cls, rag_cls)
            return rag_cls

        return decorator

    @classmethod
    def create(cls, name: RAGMethod, config: BaseRAGConfig) -> "BaseRAG":
        """Instantiate a registered RAG class using a validated config object.

        Args:
            name: Registered RAG method name.
            config: Instance of a subclass of BaseRAGConfig.

        Returns:
            Instance of the specified RAG class.

        Raises:
            KeyError: If the method name is not registered.
            TypeError: If config type does not match registered config class.

        """
        if name not in cls.registry:
            raise KeyError(f"RAG method '{name}' is not registered.")
        config_cls, rag_cls = cls.registry[name]
        if not isinstance(config, config_cls):
            raise TypeError(
                f"Expected config of type {config_cls.__name__}, got {type(config).__name__}"
            )
        return rag_cls(config)



from encourage.rag.base.enum import RAGMethod
from encourage.rag.base.interface import RAGMethodInterface
from encourage.rag.base_impl import BaseRAG
from encourage.rag.hybrid_bm25 import HybridBM25RAG
from encourage.rag.hyde import HydeRAG
from encourage.rag.known_context import KnownContext
from encourage.rag.reranker import RerankerRAG
from encourage.rag.self_rag import SelfRAG
from encourage.rag.summarize import SummarizationContextRAG, SummarizationRAG

__all__ = [
    "HydeRAG",
    "KnownContext",
    "BaseRAG",
    "RAGMethod",
    "SummarizationRAG",
    "RAGMethodInterface",
    "SummarizationContextRAG",
    "RerankerRAG",
    "HybridBM25RAG",
    "SelfRAG",
]



"""InferenceRunner class for running models."""

import json
import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Type

from openai import OpenAI
from openai.types.chat.chat_completion import (
    ChatCompletion,
    Choice,
)
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.completion_usage import CompletionUsage
from pydantic import BaseModel
from tqdm import tqdm

from encourage.llm.response_wrapper import ResponseWrapper
from encourage.prompts.prompt import Prompt
from encourage.prompts.prompt_collection import PromptCollection

if TYPE_CHECKING:
    from vllm import SamplingParams


class InferenceRunner(ABC):
    """Abstract base class for model inference."""

    def __init__(
        self,
        sampling_parameters: "SamplingParams",
        model_name: str,
        base_url: str = "http://localhost:18123/v1/",
        env_var_name: str = "VLLM_API_KEY",
    ):
        self.sampling_parameters = sampling_parameters
        self.model_name = model_name
        self.base_url = base_url
        self.api_key = os.getenv(env_var_name) or ""
        if not self.api_key:
            raise ValueError("API key cannot be empty")

    @abstractmethod
    def run(self, *args: Any, **kwargs: Any) -> ChatCompletion | ResponseWrapper:
        """Abstract method to run the model with the given inputs."""
        pass


class ChatInferenceRunner(InferenceRunner):
    """Class for single chat inference."""

    def __init__(
        self,
        sampling_parameters: "SamplingParams",
        model_name: str,
        base_url: str = "http://localhost:18123/v1/",
        env_var_name: str = "VLLM_API_KEY",
    ):
        super().__init__(sampling_parameters, model_name, base_url, env_var_name)
        self.client = OpenAI(base_url=self.base_url, api_key=self.api_key)

    def run(
        self,
        prompt: Prompt,
        response_format: Type[BaseModel] | None = None,
        raw_output: bool = False,
    ) -> ChatCompletion | ResponseWrapper:
        """Run the model with the given query."""
        if isinstance(prompt, PromptCollection):  # type: ignore
            raise ValueError("PromptCollection is not supported for single chat inference")

        extra_body = {}
        if response_format:
            extra_body = {"guided_json": response_format.model_json_schema()}

        response = self.client.chat.completions.create(
            model=self.model_name,
            messages=prompt.conversation.dialog,  # type: ignore
            max_tokens=self.sampling_parameters.max_tokens,
            temperature=self.sampling_parameters.temperature,
            top_p=self.sampling_parameters.top_p,
            extra_body=extra_body,
        )
        if raw_output:
            return response
        return ResponseWrapper.from_prompt_collection(
            [response], PromptCollection.from_prompts([prompt])
        )


class ToolInferenceRunner(InferenceRunner):
    """Class for running tool calls."""

    def run(
        self, prompt: Prompt, tool_json: list[dict], tool_functions: list[Callable]
    ) -> ResponseWrapper:
        """Run the model with the given query and tool calls."""
        client = OpenAI(base_url=self.base_url, api_key=self.api_key)
        messages = prompt.conversation.dialog
        import mlflow
        from mlflow.entities import SpanType

        with mlflow.start_span("ToolChain", span_type=SpanType.TOOL):
            completion = client.chat.completions.create(
                model=self.model_name,
                messages=messages,  # type: ignore
                max_tokens=self.sampling_parameters.max_tokens,
                temperature=self.sampling_parameters.temperature,
                top_p=self.sampling_parameters.top_p,
                tools=tool_json,  # type: ignore
            )
            tool_call = completion.choices[0].message.tool_calls[0]  # type: ignore
            args = json.loads(tool_call.function.arguments)
            matching_function = next(
                (
                    func
                    for func in tool_functions
                    if getattr(func, "__name__", None) == tool_call.function.name
                ),
                None,
            )
            result = None
            if matching_function:
                result = matching_function(**args)
            else:
                print(f"No matching function found for {tool_call.function.name}")

            messages.append(completion.choices[0].message)  # type: ignore
            messages.append(
                {
                    "role": "tool",
                    "tool_call_id": tool_call.id,
                    "content": str(result),
                }
            )
            completion_2 = client.chat.completions.create(
                model=self.model_name,
                messages=messages,  # type: ignore
                tools=tool_json,  # type: ignore
                max_tokens=self.sampling_parameters.max_tokens,
                temperature=self.sampling_parameters.temperature,
                top_p=self.sampling_parameters.top_p,
            )
        return ResponseWrapper.from_prompt_collection(
            [completion, completion_2], PromptCollection.from_prompts([prompt, prompt])
        )


class BatchInferenceRunner(InferenceRunner):
    """Class for batch chat inference."""

    def run(
        self,
        prompt_collection: PromptCollection,
        response_format: Type[BaseModel] | str | None = None,
        batch_size: int = 50,  # Default batch size
    ) -> ResponseWrapper:
        """Run the model with the given queries."""
        from litellm import batch_completion

        extra_body = {}
        if response_format:
            if isinstance(response_format, type) and issubclass(response_format, BaseModel):
                extra_body = {"guided_json": response_format.model_json_schema()}
            if isinstance(response_format, str):
                extra_body = {"guided_json": response_format}  # type: ignore

        all_messages = [prompt.conversation.dialog for prompt in prompt_collection.prompts]
        total_samples = len(all_messages)

        # Calculate the number of batches
        num_batches = (total_samples + batch_size - 1) // batch_size
        all_responses = []

        # Process in batches with progress tracking
        with tqdm(total=total_samples, desc="Processing prompts") as bar:
            for batch_idx in range(num_batches):
                start_idx = batch_idx * batch_size
                end_idx = min((batch_idx + 1) * batch_size, total_samples)
                current_batch_size = end_idx - start_idx

                # Get current batch of messages
                batch_messages = all_messages[start_idx:end_idx]
                bar.set_description(
                    f"Batch {batch_idx + 1}/{num_batches} ({current_batch_size} prompts)"
                )

                # Process the batch
                batch_responses = batch_completion(
                    model=f"hosted_vllm/{self.model_name}",
                    messages=batch_messages,
                    max_tokens=self.sampling_parameters.max_tokens,
                    base_url=self.base_url,
                    api_key=self.api_key,
                    temperature=self.sampling_parameters.temperature,
                    top_p=self.sampling_parameters.top_p,
                    extra_body=extra_body,
                )

                all_responses.extend(batch_responses)
                bar.update(current_batch_size)

        return ResponseWrapper.from_prompt_collection(
            list(map(model_response_to_chat_completion, all_responses)), prompt_collection
        )


class OpenAIChatInferenceRunner(InferenceRunner):
    """Class for model inference."""

    def __init__(
        self,
        sampling_parameters: "SamplingParams",
        model_name: str,
    ):
        env_var_name = "OPENAI_API_KEY"
        super().__init__(sampling_parameters, model_name, env_var_name=env_var_name)
        self.client = OpenAI()

    def run(
        self,
        prompt: Prompt,
        response_format: Type[BaseModel] | str | None = None,
    ) -> ResponseWrapper:
        """Run the model with the given query."""
        response = self.client.chat.completions.create(
            model=self.model_name,
            messages=prompt.conversation.dialog,  # type: ignore
            max_tokens=self.sampling_parameters.max_tokens,
            temperature=self.sampling_parameters.temperature,
            response_format=response_format,  # type: ignore
        )
        return ResponseWrapper.from_prompt_collection(
            [response], PromptCollection.from_prompts([prompt])
        )


def model_response_to_chat_completion(model_response: Any) -> ChatCompletion:
    """Convert a ModelResponse object to a ChatCompletion object."""
    from litellm.types.utils import ModelResponse

    if not isinstance(model_response, ModelResponse):
        # Handle case where response is not a ModelResponse
        return ChatCompletion(
            id="",
            choices=[],
            created=0,
            model="",
            object="chat.completion",
            system_fingerprint="",
            usage=CompletionUsage(
                completion_tokens=0,
                prompt_tokens=0,
                total_tokens=0,
                completion_tokens_details=None,
                prompt_tokens_details=None,
            ),
        )
    else:
        # For valid ModelResponse objects
        return ChatCompletion(
            id=model_response.id,
            choices=[
                Choice(
                    finish_reason="length",
                    index=choice.index,
                    logprobs=None,
                    message=ChatCompletionMessage(
                        content=choice.message.content,  # type: ignore
                        refusal=None,
                        role=choice.message.role,  # type: ignore
                        audio=None,
                        function_call=choice.message.function_call,  # type: ignore
                        tool_calls=choice.message.tool_calls or [],  # type: ignore
                    ),
                )
                for choice in model_response.choices
            ],
            created=model_response.created,
            model=model_response.model,  # type: ignore
            object=model_response.object,  # type: ignore
            service_tier=model_response.service_tier,  # type: ignore
            system_fingerprint=model_response.system_fingerprint,
            usage=CompletionUsage(
                completion_tokens=model_response.usage.completion_tokens,  # type: ignore
                prompt_tokens=model_response.usage.prompt_tokens,  # type: ignore
                total_tokens=model_response.usage.total_tokens,  # type: ignore
                completion_tokens_details=model_response.usage.completion_tokens_details,  # type: ignore
                prompt_tokens_details=model_response.usage.prompt_tokens_details,  # type: ignore
            ),
        )



"""Module that defines the ResponseWrapper class."""

import logging
from typing import Iterator, Optional

from openai.types.chat import ChatCompletion, ChatCompletionMessage

from encourage.llm.response import Response
from encourage.prompts.context import Context
from encourage.prompts.conversation import Conversation, Role
from encourage.prompts.meta_data import MetaData
from encourage.prompts.prompt import Prompt
from encourage.prompts.prompt_collection import PromptCollection
from encourage.utils.tracing import enable_tracing

# Configure logging
logger = logging.getLogger(__name__)


class ResponseWrapper:
    """Class that aggregates ChatCompletion with corresponding PromptCollection details."""

    @classmethod
    def from_prompt_collection(
        cls,
        chat_completions: list[ChatCompletion],
        collection: PromptCollection,
    ) -> "ResponseWrapper":
        """Create ResponseWrapper from ChatCompletion and PromptCollection or Conversation."""
        if len(chat_completions) != len(collection.prompts):
            raise ValueError("The number of request outputs does not match the number of prompts.")

        # Create responses for each chat_completion and its corresponding prompt
        responses = [
            cls.handle_prompt_response(chat_completion, prompt)
            for chat_completion, prompt in zip(chat_completions, collection.prompts)
        ]
        return cls(responses)

    @classmethod
    @enable_tracing(span_name="handle_prompt_response")
    def handle_prompt_response(cls, chat_completion: ChatCompletion, prompt: Prompt) -> Response:
        """Create a Response object from a ChatCompletion and Prompt."""
        prompt.conversation = transform_chat_completion_to_conversation(prompt.conversation)

        return Response(
            request_id=chat_completion.id,
            prompt_id=str(prompt.id),
            conversation_id=0,
            sys_prompt=prompt.conversation.sys_prompt,
            user_prompt=prompt.conversation.get_last_message_by_user(),
            response=chat_completion.choices[0].message.content
            if chat_completion.choices
            else "No response",
            meta_data=prompt.meta_data,
            context=prompt.context,
            arrival_time=(chat_completion.created if chat_completion.created is not None else 0.0),
            finished_time=0.0,
        )

    @classmethod
    def from_conversation(
        cls,
        chat_completions: list[ChatCompletion],
        conversation: Conversation,
        contexts: list[Context] = [],
        meta_datas: list[MetaData] = [],
    ) -> "ResponseWrapper":
        """Create ResponseWrapper from ChatCompletion and Conversation."""
        # Check for mismatched lengths
        if len(chat_completions) != len(conversation.get_messages_by_role(Role.USER)):
            raise ValueError("The number of request outputs does not match the number of messages.")

        # Safely access the system prompt, handle missing system messages
        sys_messages = conversation.get_messages_by_role(Role.SYSTEM)
        if not sys_messages:
            raise ValueError("No system messages found in the conversation.")
        sys_prompt = sys_messages[0]["content"]

        # Extract user messages
        user_messages = [msg["content"] for msg in conversation.get_messages_by_role(Role.USER)]

        responses = []
        meta_datas = meta_datas or [MetaData()] * len(chat_completions)
        contexts = contexts or [Context()] * len(chat_completions)
        for conversation_id, (chat_completion, user_message, meta_data, context) in enumerate(
            zip(chat_completions, user_messages, meta_datas, contexts)
        ):
            response = cls.handle_conversation_response(
                sys_prompt, conversation_id, chat_completion, user_message, context, meta_data
            )
            responses.append(response)

        return cls(responses)

    @staticmethod
    @enable_tracing(span_name="handle_conversation_response")
    def handle_conversation_response(
        sys_prompt: str,
        conversation_id: int,
        chat_completion: ChatCompletion,
        message: str,
        context: Context,
        meta_data: MetaData,
    ) -> Response:
        """Create a Response object from a ChatCompletion and Conversation."""
        return Response(
            request_id=chat_completion.id,
            prompt_id="",
            conversation_id=conversation_id,
            sys_prompt=sys_prompt,
            user_prompt=message,
            response=chat_completion.choices[0].message.content
            if chat_completion.choices
            else "No response",
            context=context,
            meta_data=meta_data,
            arrival_time=(chat_completion.created if chat_completion.created is not None else 0.0),
            finished_time=0.0,
        )

    def get_responses(self) -> list[str]:
        """Return the list of actually text responses from the model."""
        return [response.response for response in self.response_data]

    def print_response_summary(self) -> None:
        """Prints a summary of all responses."""
        for response in self.response_data:
            response.print_response()

    def print_response_by_prompt_id(self, prompt_id: str) -> None:
        """Print the response details for a specific prompt ID."""
        response = self._find_response_by("prompt_id", prompt_id)
        if response:
            response.print_response()
        else:
            print(f"No response found for Prompt ID: {prompt_id}")

    def print_response_by_request_id(self, request_id: str) -> None:
        """Print the response details for a specific request ID."""
        response = self._find_response_by("request_id", request_id)
        if response:
            response.print_response()
        else:
            logger.error(f"No response found for Request ID: {request_id}")

    def get_response_by_prompt_id(self, prompt_id: str) -> Optional[Response]:
        """Return the response details for a specific prompt ID."""
        return self._find_response_by("prompt_id", prompt_id)

    def get_response_by_request_id(self, request_id: str | int) -> Response | None:
        """Return the response details for a specific request ID."""
        return self._find_response_by("request_id", str(request_id))

    def _find_response_by(self, key: str, value: str) -> Response | None:
        """Helper method to find a response by a given key and value."""
        for response in self.response_data:
            if getattr(response, key) == value:
                return response
        return None

    def __repr__(self) -> str:
        return f"ResponseWrapper({self.response_data})"

    def __init__(self, responses: list[Response]):
        self.response_data = responses

    def __iter__(self) -> Iterator[Response]:
        """Allows iteration over the responses."""
        return iter(self.response_data)

    def __getitem__(self, key: int) -> Response:
        """Returns the response at the given index."""
        if not self.response_data:
            raise IndexError("Response data is empty.")
        if key < 0 or key >= len(self.response_data):
            raise IndexError("Index out of range.")
        return self.response_data[key]

    def __len__(self) -> int:
        """Returns the number of responses."""
        return len(self.response_data)


def transform_chat_completion_to_conversation(conversation: Conversation) -> Conversation:
    """Transforms a Conversation object to a ChatCompletion object."""
    # Create a new conversation object
    new_dialog = []
    for message in conversation.dialog:
        if isinstance(message, ChatCompletionMessage):  # type: ignore
            message = {"role": Role.TOOL, "content": message.tool_calls[0].function.arguments}  # type: ignore
        new_dialog.append(message)
    conversation.dialog = new_dialog
    return conversation
